// This is a roxen module. Copyright  2000 - 2004, Roxen IS.
//

#include <module.h>
inherit "module";

#define _ok RXML_CONTEXT->misc[" _ok"]

constant cvs_version = "$Id: additional_rxml.pike,v 1.38 2008/02/12 16:08:28 stewa Exp $";
constant thread_safe = 1;
constant module_type = MODULE_TAG;
constant module_name = "Tags: Additional RXML tags";
constant module_doc  = "This module provides some more complex and not as widely used RXML tags.";

void create() {
  defvar("insert_href",
#ifdef THREADS
	 1
#else
	 0
#endif
	 , "Allow <insert href>",
	 TYPE_FLAG, #"\
If set, it will be possible to use <tt>&lt;insert href&gt;</tt> to
insert pages from another web server. Note that the handler thread
will be blocked while it fetches the web page, so if enough requests
are made to an unresponding external web server, this server might
also cease to respond.");

  defvar ("default_timeout", 300, "Default timeout for <insert href>",
	  TYPE_INT, #"\
Maximum waiting time in seconds for a response from the other web
server when <tt>&lt;insert href&gt;</tt> is used. This can be
overridden with the <tt>timeout</tt> attribute. Set it to zero to wait
indefinitely by default.

<p>Timeout is not available if your run the server without threads."
#ifndef THREADS
	  " <strong>You are currently running without threads.</strong>"
#endif
	 );

  defvar("recursion_limit", 2, "Maximum recursion depth for <insert href>",
	 TYPE_INT|VAR_MORE,
	 "Maxumum number of nested <tt>&lt;insert href&gt;</tt>'s allowed. "
	 "May be set to zero to disable the limit.");
}

#ifdef THREADS

class AsyncHTTPClient {
  int status;
  object con;
  Standards.URI url;
  string path, query, req_data,method;
  mapping request_headers;

  Thread.Queue queue = Thread.Queue();

  void do_method(string _method,
		 string|Standards.URI _url,
		 void|mapping post_variables,
		 void|mapping _request_headers,
		 void|Protocols.HTTP.Query _con, void|string _data)
  {
    if(!_con) {
      con = Protocols.HTTP.Query();
    }
    else
      con = _con;

    method = _method;

    if(!_request_headers)
      request_headers = ([]);
    else
      request_headers = _request_headers;

    if(post_variables) {
      request_headers += 
		   (["content-type":
		     "application/x-www-form-urlencoded"]);
      _data = Protocols.HTTP.http_encode_query(post_variables);
    }

    req_data = _data;

    if(stringp(_url)) {
      if (mixed err = catch (url=Standards.URI(_url)))
	RXML.parse_error ("Invalid URL: %s\n", describe_error (err));
    }
    else
      url = _url;

#if constant(SSL.sslfile) 	
    if(url->scheme!="http" && url->scheme!="https")
      error("Protocols.HTTP can't handle %O or any other protocols than HTTP or HTTPS\n",
	    url->scheme);
    
    con->https= (url->scheme=="https")? 1 : 0;
#else
    if(url->scheme!="http"	)
      error("Protocols.HTTP can't handle %O or any other protocol than HTTP\n",
	    url->scheme);
    
#endif
    
    if(!request_headers)
      request_headers = ([]);
    mapping default_headers = ([
      "user-agent" : "Mozilla/4.0 compatible (Pike HTTP client)",
      "host" : sprintf("%s:%d", url->host, url->port) ]);
    
    if(url->user || url->passwd)
      default_headers->authorization = "Basic "
	+ MIME.encode_base64(url->user + ":" +
			     (url->password || ""));
    request_headers = default_headers | request_headers;
    
    query=url->query;
    /*
    if(query_variables && sizeof(query_variables))
      {
	if(query)
	  query+="&"+Protocols.HTTP.http_encode_query(query_variables);
	else
	  query=Protocols.HTTP.http_encode_query(query_variables);
      }
    */

    
    path=url->path;
    if(path=="") path="/";
  }


  string data() {
    if(!con->ok)
      return 0;
    return con->data();
  }
 
  void req_ok() {
    status = con->status;
    // Wake up handler thread
    queue->write("@");
  }
  
  void req_fail() {
    //werror("Insert HREF request failed or timed out.\n");
    status = 0;
    // Wake up handler thread
    queue->write("@");
  }


  void run() {
    /*
    con->sync_request(url->host,url->port,
		      method+" "+path+(query?("?"+query):"")+" HTTP/1.0",
		      request_headers, req_data);
    
    */
    con->set_callbacks(req_ok, req_fail);
    con->async_request(url->host,url->port,
		       method+" "+path+(query?("?"+query):"")+" HTTP/1.0",
		       request_headers, req_data);
    status = con->status;
    queue->read();
  }

  void create(string method, mapping args, mapping|void headers) {
    if(method == "POST") {
      mapping vars = ([ ]);
#if constant(roxen)
      foreach( (args["post-variables"] || "") / ",", string var) {
	array a = var / "=";
	if(sizeof(a) == 2)
	  vars[String.trim_whites(a[0])] = RXML.user_get_var(String.trim_whites(a[1]));
      }
#endif
      do_method("POST", args->href, vars, headers);
    }
    else
      do_method("GET", args->href, 0, headers);

    if(args->timeout)
      con->timeout = (int) args->timeout;
    else if (int t = global::query ("default_timeout"))
      con->timeout = t;
    else
      // There ought to be a way to disable the timeout in
      // Protocols.HTTP.Query.
      con->timeout = 2147483647;
  }
  
}
#endif

class TagInsertHref {
  inherit RXML.Tag;
  constant name = "insert";
  constant plugin_name = "href";

  string get_data(string var, mapping args, RequestID id) {
    if(!query("insert_href")) RXML.run_error("Insert href is not allowed.");

    int recursion_depth = (int)id->request_headers["x-roxen-recursion-depth"];

    if (query("recursion_limit") &&
	(recursion_depth >= query("recursion_limit")))
      RXML.run_error("Too deep insert href recursion.");

    recursion_depth++;

    if(args->nocache)
      NOCACHE();
    else
      CACHE(60);

    string method = "GET";
    if(args->method && lower_case(args->method) == "post")
      method = "POST";

    object /*Protocols.HTTP|AsyncHTTPClient*/ q;

    mapping(string:string) headers = ([ "X-Roxen-Recursion-Depth":
					(string)recursion_depth ]);
    if (args["request-headers"])
      foreach (args["request-headers"] / ",", string header)
	if (sscanf (header, "%[^=]=%s", string name, string val) == 2)
	  headers[name] = val;

#ifdef THREADS
    q = AsyncHTTPClient(method, args, headers);
    q->run();
#else

    mixed err;
    if(method == "POST") {
      mapping vars = ([ ]);
      foreach( (args["post-variables"] || "") / ",", string var) {
	array a = var / "=";
	if(sizeof(a) == 2)
	  vars[String.trim_whites(a[0])] = RXML.user_get_var(String.trim_whites(a[1]));
      }
      err = catch {
	  q = Protocols.HTTP.post_url(args->href, vars, headers);
	};
    }
    else
      err = catch {
	  q = Protocols.HTTP.get_url(args->href, 0, headers);
	};
    if (err) {
      string msg = describe_error (err);
      if (has_prefix (msg, "Standards.URI:"))
	RXML.parse_error ("Invalid URL: %s\n", msg);
    }
#endif
    
    _ok = 1;
    
    if(args["status-variable"] && q && q->status)
      RXML.user_set_var(args["status-variable"],q->status);
    
    if(q && q->status>0 && q->status<400) {
      string data =  q->data();

      function get_ct_charset =
	lambda(string ct) {
	  string cs;
	  foreach((ct/";")[1..], string s) {
	    string s2 = String.trim_all_whites(s);
	    string _cs;
	    if(sscanf(s2, "charset=%s", _cs) == 1)
	      cs = String.trim_all_whites(_cs);
	  }
	  return cs;
	};
      
      if(string ct = q->con->headers["content-type"]) {
	ct = String.trim_all_whites(ct);
	string cs;
	if(has_prefix(ct,"text/")) {
	  cs = get_ct_charset(ct);
	  if(!cs && has_prefix(ct,"text/html")) {
	    Parser.HTML parser = Parser.HTML();
	    parser->case_insensitive_tag(1);
	    parser->lazy_entity_end(1);
	    parser->ignore_unknown(1);
	    parser->match_tag(0);
	    parser->add_tags( ([ "meta": lambda( Parser.HTML p, mapping m) 
					 {
					   if(m["content"] && m["http-equiv"] && 
					      lower_case(m["http-equiv"]) == "content-type")
					     cs = get_ct_charset(m["content"]);
					 } ]) );
	    parser->finish(data);
	  }
	}
	
	if(cs) {
	  catch {
	    data = Locale.Charset.decoder(cs)->feed(data)->drain();
	  };
	}

      }
      
      return data;
    }    

    _ok = 0;

    if(!args->silent)
      RXML.run_error((q && q->status_desc) || "No server response");
    return "";
  }
}

string container_recursive_output (string tagname, mapping args,
                                  string contents, RequestID id)
{
  int limit;
  array(string) inside, outside;
  if (id->misc->recout_limit)
  {
    limit = id->misc->recout_limit - 1;
    inside = id->misc->recout_outside, outside = id->misc->recout_inside;
  }
  else
  {
    limit = (int) args->limit || 100;
    inside = args->inside ? args->inside / (args->separator || ",") : ({});
    outside = args->outside ? args->outside / (args->separator || ",") : ({});
    if (sizeof (inside) != sizeof (outside))
      RXML.parse_error("'inside' and 'outside' replacement sequences "
		       "aren't of same length.\n");
  }

  if (limit <= 0) return contents;

  int save_limit = id->misc->recout_limit;
  string save_inside = id->misc->recout_inside, save_outside = id->misc->recout_outside;

  id->misc->recout_limit = limit;
  id->misc->recout_inside = inside;
  id->misc->recout_outside = outside;

  string res = Roxen.parse_rxml (
    parse_html (
      contents,
      (["recurse": lambda (string t, mapping a, string c) {return c;}]),
      ([]),
      "<" + tagname + ">" + replace (contents, inside, outside) +
      "</" + tagname + ">"),
    id);

  id->misc->recout_limit = save_limit;
  id->misc->recout_inside = save_inside;
  id->misc->recout_outside = save_outside;

  return res;
}

class TagSprintf {
  inherit RXML.Tag;
  constant name = "sprintf";

  class Frame {
    inherit RXML.Frame;

    array do_return(RequestID id) {
      array(mixed) in;
      if(args->split)
	in=(content || "")/args->split;
      else
	in=({content || ""});

      array f=((args->format-"%%")/"%")[1..];
      if(sizeof(in)!=sizeof(f))
	RXML.run_error("Indata hasn't the same size as format data (%d, %d).\n", sizeof(in), sizeof(f));

      // Do some casting
      for(int i; i<sizeof(in); i++) {
	int quit;
	foreach(f[i]/1, string char) {
	  if(quit) break;
	  switch(char) {
	  case "d":
	  case "u":
	  case "o":
	  case "x":
	  case "X":
	  case "c":
	  case "b":
	    in[i]=(int)in[i];
	    quit=1;
	    break;
	  case "f":
	  case "g":
	  case "e":
	  case "G":
	  case "E":
	  case "F":
	    in[i]=(float)in[i];
	    quit=1;
	    break;
	  case "s":
	  case "O":
	  case "n":
	  case "t":
	    quit=1;
	    break;
	  }
	}
      }

      result=sprintf(args->format, @in);
      return 0;
    }
  }
}

class TagSscanf {
  inherit RXML.Tag;
  constant name = "sscanf";
  mapping(string:RXML.Type) req_arg_types = ([ "variables" : RXML.t_text(RXML.PEnt),
					       "format"    : RXML.t_text(RXML.PEnt)
  ]);
  mapping(string:RXML.Type) opt_arg_types = ([ "return"    : RXML.t_text(RXML.PEnt),
					       "scope"     : RXML.t_text(RXML.PEnt)
  ]);

  class Frame {
    inherit RXML.Frame;

    string do_return(RequestID id) {
      array(string) vars=args->variables/",";
      array(string) vals;
      mixed err = catch {
	  vals = array_sscanf(content || "", args->format);
	};
      if(err) {
        string msg = "Unknown error.\n";
        if(arrayp(err) && sizeof(err) && stringp(err[0]))
          msg = err[0];
        RXML.run_error(msg);
      }
      if(sizeof(vars)<sizeof(vals))
	RXML.run_error("Too few variables.\n");

      int var=0;
      foreach(vals, string val)
	RXML.user_set_var(vars[var++], val, args->scope);

      if(args->return)
	RXML.user_set_var(args->return, sizeof(vals), args->scope);
      return 0;
    }
  }
}

class TagDice {
  inherit RXML.Tag;
  constant name = "dice";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame {
    inherit RXML.Frame;

    string do_return(RequestID id) {
      NOCACHE();
      if(!args->type) args->type="D6";
      args->type = replace( args->type, "T", "D" );
      int value;
      args->type=replace(args->type, "-", "+-");
      foreach(args->type/"+", string dice) {
	if(has_value(dice, "D")) {
	  if(dice[0]=='D')
	    value+=random((int)dice[1..])+1;
	  else {
	    array(int) x=(array(int))(dice/"D");
	    if(sizeof(x)!=2)
	      RXML.parse_error("Malformed dice type.\n");
	    value+=x[0]*(random(x[1])+1);
	  }
	}
	else
	  value+=(int)dice;
      }

      if(args->variable)
	RXML.user_set_var(args->variable, value, args->scope);
      else
	result=(string)value;

      return 0;
    }
  }
}

class TagEmitKnownLangs
{
  inherit RXML.Tag;
  constant name = "emit", plugin_name = "known-langs";
  array get_dataset(mapping m, RequestID id)
  {
    return map(roxenp()->list_languages(),
	       lambda(string id)
	       {
		 object language = roxenp()->language_low(id);
		 string eng_name = language->id()[1];
		 if(eng_name == "standard")
		   eng_name = "english";
		 return ([ "id" : id,
			 "name" : language->id()[2],
		  "englishname" : eng_name ]);
	       });
  }
}

class TagFormatNumber
{
  inherit RXML.Tag;
  constant name  = "format-number";

  mapping(string:RXML.Type) req_arg_types = 
    ([ /* "value"                : RXML.t_text(RXML.PEnt), */
       "pattern"              : RXML.t_text(RXML.PEnt)
    ]);
  mapping(string:RXML.Type) opt_arg_types = 
    ([ "decimal-separator"    : RXML.t_text(RXML.PEnt),
       "group-separator"      : RXML.t_text(RXML.PEnt)
    ]); 

  class Frame {
    inherit RXML.Frame;

    string do_return(RequestID req_id) {
      constant PERCENT   = 0x01;
      constant PER_MILLE = 0x02;
      constant TRUE      = 0x01;
      constant FALSE     = 0x00;
      constant FRAC_PART = 1;
      constant INT_PART  = 0;
      
      string grp_sep     = ",";
      string dec_sep     = "."; 
      if (args["group-separator"])
      {
	grp_sep = args["group-separator"];
      }
      if (args["decimal-separator"])
      {
	dec_sep = args["decimal-separator"];
      }

      // Parse the number.
      array(string) value = ({"", ""});
      int int_part = 0, frac_part = 0;
      if (sscanf(content, "%*s%d.%d", int_part, frac_part) == 3)
      {
	value[INT_PART]  = (string) int_part;
	value[FRAC_PART] = (string) frac_part;
      }
      else if (sscanf(content, "%*s-.%d", frac_part) == 2)
      {
	value[INT_PART]  = "-0";
	value[FRAC_PART] = (string) frac_part; 
      }
      else if (sscanf(content, "%*s.%d", frac_part) == 2)
      {
	value[INT_PART]  = "0";
	value[FRAC_PART] = (string) frac_part;
      }
      else if (sscanf(content, "%*s%d", int_part) == 2)
      {
	value[INT_PART]  = (string) int_part;
	value[FRAC_PART] = "";
      }
      else
      {
	RXML.parse_error("No number to be formatted was given.");
      }


      /* We need to break down the pattern, first by ";" and make sure that ";"
         is not an escaped character. */
      array(string) temp_pattern = args->pattern / ";";
      array(string) pattern = ({ "", "" });
      int is_pos = TRUE;
      for (int i = 0; i < sizeof(temp_pattern); i++)
      {
	pattern[is_pos] += temp_pattern[i];
	if (sizeof(pattern[is_pos]) > 0 &&
	    pattern[is_pos][sizeof(pattern[is_pos]) - 1] == "'"[0])
	{
	  pattern[is_pos][sizeof(pattern[is_pos]) - 1] = ";"[0];
	}
	else if (is_pos)
	{
	    is_pos = FALSE;
	}
      }

      /* Now that we have extracted the patterns for positive and negative 
         numbers, we can reuse the is_pos flag to reflect which pattern we
	 should use. */
      if ((((int)value[INT_PART]) < 0 || value[INT_PART][0..0] == "-") && 
	  sizeof(pattern[0]) > 0)
      {
	is_pos = FALSE;
	// Trim away the minus sign.
	sscanf(value[0], "%*[ ]-%s", value[0]) ;
      }
      else if (sizeof(pattern[1]) > 0)
      {
	is_pos = TRUE;
      }
      else
      {
	RXML.parse_error("The value was positive but no pattern " +
			 "for positive values was defined.\n");
      }
      
      // Now do the same thing as above to find the fraction delimiter.
      temp_pattern = pattern[is_pos] / ".";
      pattern = ({"", ""});
      int is_frac = FALSE;

      pattern[INT_PART] = temp_pattern[0];
      if (sizeof(pattern[INT_PART]) > 0 &&
	  pattern[INT_PART][-1] == "'"[0])
      {
	pattern[INT_PART][sizeof(pattern[INT_PART]) - 1] = "."[0];
      }
      else
      {
	is_frac = TRUE;
      }

      if (sizeof(temp_pattern) > 1)
      {
	for (int i = 1; i < sizeof(temp_pattern); i++)
	{
	  pattern[is_frac] += temp_pattern[i];
	  if (sizeof(pattern[is_frac]) > 0 &&
	      pattern[is_frac][-1] == "'"[0])
	  {
	    pattern[is_frac][-1] = "."[0];
	  }
	  else if (!is_frac)
	  {
	    is_frac = TRUE;
	  }
	  else if (i < (sizeof(temp_pattern) - 1))
	  {
	    pattern[is_frac] += ".";
	  }
	}
      }

      // Handle percent and per-mille
      int log_ten = 0;
      if (search(pattern[FRAC_PART], "%") >= 0 ||
	  search(pattern[INT_PART], "%") >= 0)
      {
	log_ten = 2;
      }
      else if (search(pattern[FRAC_PART], "\x2030") >= 0 ||
	       search(pattern[INT_PART], "\x2030") >= 0)
      {
	log_ten = 3;
      }

      for (int i = log_ten; i > 0; i--)
      {
	if (strlen(value[FRAC_PART]) > 0)
	{
	  value[INT_PART] += value[FRAC_PART][0..0];
	  value[FRAC_PART] = value[FRAC_PART][1..];
	}
        else 
	{
	  value[INT_PART] += "0";
	}
      }

      // Trim away leading zeros
      value[INT_PART] = (string)(int)value[INT_PART][0..];

      // Start with the fractional part.
      int val_length     = strlen(value[FRAC_PART]);
      int ptn_length     = strlen(pattern[FRAC_PART]);
      int vi             = 0;
      string frac_result = "";
      
      foreach(pattern[FRAC_PART]/1, string s)
      {
	switch(s)
	{
	  case "'":
	    break;
	  case "#":
	    if (vi < val_length)
	    {
	      frac_result += value[FRAC_PART][vi..vi];
	      vi++;
	    }
	    break;
	  case "0":
	    if (vi < val_length)
	    {
	      frac_result += value[FRAC_PART][vi..vi];
	      vi++;
	    } 
	    else
	    {
	      frac_result += "0";
	    }
	    break;
	  default:
	    frac_result += s;
	}
      }

      // Round the last digit.    
      if (vi < val_length &&
	  ((int)value[FRAC_PART][vi..vi] > 5 ||
	  ((int)value[FRAC_PART][vi..vi] == 5 && is_pos)))
      {
	if (strlen(frac_result) == 0)
	{
	  value[INT_PART] = (string)(((int)value[INT_PART]) + 1);  
	}
	else
	{
	  frac_result[-1] = frac_result[-1] + 1;  
	}
      }
      
      // Now for the integral part. We traverse it in reverse.
      val_length        = strlen(value[INT_PART]);
      ptn_length        = strlen(pattern[INT_PART]);
      vi                = 0;
      int num_wid       = 0; 
      string rev_val    = reverse(value[INT_PART]);
      string int_result = "";
      string minus_sign = "";
      foreach(reverse(pattern[INT_PART])/1, string s)
      {
	switch(s)
	{
	  case "'":
	    break;
	  case "#":
	    if (vi < val_length)
	    {
	      int_result += rev_val[vi..vi];
	      vi++;
	    }
	    if (num_wid <= 0)
	    {
	      num_wid--;
	    }
	    break;
	  case "0":
	    if (vi < val_length)
	    {
	      int_result += rev_val[vi..vi];
	      vi++;
	    } 
	    else
	    {
	      int_result += "0";
	    }
	    if (num_wid <= 0)
	    {
	      num_wid--;
	    }
	    break;
	  case ",":
	    if (num_wid <= 0)
	    {
	      num_wid *= -1;
	    }
	    if (vi < val_length)
	    {
	      int_result += grp_sep;
	    }
	    break;
	  case "-":
	    if (is_pos)
	    {
	      int_result += s;
	    }
	    else if (minus_sign = "")
	    {
	      minus_sign = "-";
	    }
	    break;
	  case "%":
	  case "\x2030":
	    frac_result = s;
	    break;
	  default:
	    int_result += s;
	}
      }
      for (;vi < val_length;vi++)
	{
	  if (num_wid > 0 && vi%num_wid == 0)
	  {
	    int_result += grp_sep;
	  }
	  int_result += rev_val[vi..vi];
      }

      if (strlen(frac_result) == 0)
      {
	result += minus_sign + reverse(int_result);
      }
      else if (frac_result[0..0] == "%" ||
	       frac_result[0..0] == "\x2030")
      {
	result += minus_sign + reverse(int_result) + frac_result[0..0];
      }
      else 
      {
	result += minus_sign + reverse(int_result) + dec_sep + frac_result;
      }
    }
  } 
}



TAGDOCUMENTATION;
#ifdef manual
constant tagdoc=([
  "dice":#"<desc type='tag'><p><short>
 Simulates a D&amp;D style dice algorithm.</short></p></desc>

<attr name='type' value='string' default='D6'><p>
 Describes the dices. A six sided dice is called 'D6' or '1D6', while
 two eight sided dices is called '2D8' or 'D8+D8'. Constants may also
 be used, so that a random number between 10 and 20 could be written
 as 'D9+10' (excluding 10 and 20, including 10 and 20 would be 'D11+9').
 The character 'T' may be used instead of 'D'.</p>
</attr>",

  "insert#href":#"<desc type='plugin'><p><short>
 Inserts the contents at that URL.</short> This function has to be
 enabled in the <module>Additional RXML tags</module> module in the
 Roxen WebServer configuration interface. The page download will block
 the current thread, and if running unthreaded, the whole server.
 There is no timeout in the download, so if the server connected to
 hangs during transaction, so will the current thread in this server.</p></desc>

<attr name='href' value='string'><p>
 The URL to the page that should be inserted.</p>
</attr>

<attr name='nocache' value='string'><p>
 If provided the resulting page will get a zero cache time in the RAM cache.
 The default time is up to 60 seconds depending on the cache limit imposed by
 other RXML tags on the same page.</p>
</attr>

<attr name='method' value='string' default='GET'><p>
 Method to use when requesting the page. GET or POST.</p>
</attr>

<attr name='silent' value='string' ><p>
 Do not print any error messages.</p>
</attr>

<attr name='timeout' value='int' ><p>
 Timeout for the request in seconds. This is not available if your run the server without threads.</p>
<ex-box>
<insert href='http://www.somesite.com/news/' silent='silent' timeout='10' />
<else>
   Site did not respond within 10 seconds. Try again later.
</else>
</ex-box>
</attr>

<attr name='post-variables' value='\"variable=rxml variable[,variable2=rxml variable2,...]\"'><p>
 Comma separated list of variables to send in a POST request.</p>
<ex-box>
<insert href='http://www.somesite.com/news/' 
         method='POST' post-variables='action=var.action,data=form.data' />
</ex-box> 
</attr>

<attr name='request-headers' value='\"header=value[,header2=value2,...]\"'><p>
 Comma separated list of extra headers to send in the request.</p>
</attr>
",

"sscanf":#"<desc type='cont'><p><short>
 Extract parts of a string and put them in other variables.</short> Refer to
 the sscanf function in the Pike reference manual for a complete
 description.</p>
</desc>

<attr name='variables' value='list' required='required'><p>
 A comma separated list with the name of the variables that should be set.</p>
<ex>
<sscanf variables='form.year,var.month,var.day'
format='%4d%2d%2d'>19771003</sscanf>
&form.year;-&var.month;-&var.day;
</ex>
</attr>

<attr name='scope' value='name' required='required'><p>
 The name of the fallback scope to be used when no scope is given.</p>
<ex>
<sscanf variables='year,month,day' scope='var'
 format='%4d%2d%2d'>19801228</sscanf>
&var.year;-&var.month;-&var.day;<br />
<sscanf variables='form.year,var.month,var.day'
 format='%4d%2d%2d'>19801228</sscanf>
&form.year;-&var.month;-&var.day;
</ex>
</attr>

<attr name='return' value='name'><p>
 If used, the number of successful variable 'extractions' will be
 available in the given variable.</p>
</attr>",

  "sprintf":#"<desc type='cont'><p><short>
 Prints out variables with the formating functions availble in the
 Pike function sprintf.</short> Refer to the Pike reference manual for
 a complete description.</p></desc>

<attr name='format' value='string'><p>
  The formatting string.</p>
</attr>

<attr name='split' value='charater'><p>
  If used, the tag content will be splitted with the given string.</p>
<ex>
<sprintf format='#%02x%02x%02x' split=','>250,0,33</sprintf>
</ex>
</attr>",

  "format-number":#"<desc type='cont'><p><short>
Formats a number according to pattern passed as an argument. </short>
Useful for adding grouping and rounding numbers to a fixed number 
of fraction digit characters.
</p>
<p>
<ex><set variable='var.foo'>2318543.78</set>

<p><format-number pattern='0.000'>&var.foo;</format-number></p>

<p>
  <format-number pattern='#,###' 
		 group-separator=' '
		 decimal-separator=','>
    &var.foo;
  </format-number>
</p></ex>
</p>
<p>
  It's important that the guidelines for the pattern are followed. 
  Bad patterns will normally not cause an RXML-parse error, 
  instead <tag>format-number</tag> will try to format the number 
  according to the given pattern anyway.
</p>
<p>
  <tag>format-number</tag> does not handle infinity.
</p>
</desc>

<attr name='pattern' value='string' required='required'>
<p>
  The format pattern that dictates how the digits should be represented on
  screen.
</p>
<p>
  The following symbols are allowed:
</p>
<list type=\"dl\">
  <item name=\"# - Digit\">
    <p>
      Placeholder for optional digits. The number of # after the decimal 
      separator indicates number fo fraction digits to be shown.
    </p>
    <p>      
<ex><set variable='var.foo'>18543.78903</set>

<p><format-number pattern='###'>&var.foo;</format-number></p>
      
<p><format-number pattern='#.###'>&var.foo;</format-number></p></ex>
    </p>
  </item>
  <item name=\"0 - Zero Digit\">
    <p>
      Placeholder for required digits.
      If there are more digits required than given, the remaining positions
      will be filled with zeros. 
    </p>
    <p>
      0 should be placed to the right of # in the 
      pattern in the integral part and to the left of # in the fraction part.
      If they are not and there are less digits than #, the zeros will be 
      concatenated to the beginning of the integral part or the end of the
      fractional part giving unwanted results.
    </p>
    <p>
<ex><set variable='var.foo'>543.78</set>

<p><format-number pattern='#00.00'>&var.foo;</format-number></p>

<p><format-number pattern='0000.000'>&var.foo;</format-number></p>
</ex>
    </p>
  </item>
  <item name=\". - Decimal separator\">
    <p>
      Placeholder for decimal separator. If more than one are present then the
      leftmost will be used. And the remaining will be printed out \"as is\".
    </p>
    <p>
      The character used as decimal separator in the output can be changed 
      with attribute \"decimal-separator\".
    </p>
    <p>
<ex><set variable='var.foo'>543.789</set>
<p><format-number pattern='##.##'>&var.foo;</format-number></p>

<p>
  <format-number pattern='.######' decimal-separator=','>
    &var.foo;
  </format-number>
</p>
</ex>
    </p>
  </item>
  <item name=\", - Grouping separator\">
    <p>
      Placeholder for group separator. The character in output can be changed
      with attribute \"group-separator\".
    </p>
    <p>
<ex>
<p><format-number pattern='#,###'>1677518543</format-number></p>

<p><format-number pattern='#,###,###'>8543</format-number></p>
</ex>
    </p>
  </item>
  <item name=\"; - Pattern separator\">
  <p>
    Separates pattern for positive and negative result respectively. If more 
    than one ; is present, everything after the second one will be ignored.
  </p>
  <p>
<ex>
<p><format-number pattern='#,###.##;-#'>1000.50</format-number></p>
<p><format-number pattern='#,###.##;-#'>-1000.50</format-number></p>
</ex>
  </p>
  </item>
  <item name=\"% -  Percent\">
    <p>
      The number will be multiplied by 100 and a percent sign will be added
      to the output.
    </p>
    <p>
<ex>
<p><format-number pattern='#,###%'>10.50</format-number></p>

<p><format-number pattern='#.####%'>0.0003</format-number></p>
</ex>
    </p>
  </item>
  <item name=\"&#x2030; - Per-mille\">
    <p>
      The number will be multiplied by 1000 and a per-mille sign will be added
      to the output.
    </p>
    <p>
<ex>
<p><format-number pattern='##&#x2030;'>0.003</format-number></p>
</ex>
    </p>      
  </item>
  <item name=\"- - minus sign\">
    <p>
      Default negative prefix
    </p>
  </item>
  <item name=\"' - escape character\">
    <p>
      Can be used to escape characters that normally has another meaning.
    </p>
    <p>
<ex>
<p><format-number pattern=\"#'.###\">1255.34</format-number></p>
</ex>
    </p>
  </item>
</list>
</attr>

<attr name='decimal-separator' value='' default='.'>
  <p>
    A character to be used as decimal separator. Can be useful for european
    formatting.
  </p>
  <p>
<ex>
<p>
  <format-number pattern='###.##' decimal-separator=','>
    543.78
  </format-number>
</p>

<p>Even this attribute can be abused:</p>
<p>
  <format-number pattern='###.##' 
	         decimal-separator=' &lt;--integral part, fractional part--&gt; '>
    5023.643
  </format-number>
</p></ex>
  </p>
</attr>

<attr name='group-separator' value='' default=','>
  <p>
    A character to be used as group separator. Can be useful for european
    formatting.
  </p>
  <p>
<ex><set variable='var.foo'>20070901</set>
<p>
  <format-number pattern='#,###' group-separator='.'>
    &var.foo;
  </format-number>
</p>
<p>
  <format-number pattern='####,##,##' group-separator='-'>
    &var.foo; 
  </format-number>
</p>
</ex>
  </p>
</attr>",

"emit#known-langs":({ #"<desc type='plugin'><p><short>
 Outputs all languages partially supported by roxen for writing
 numbers, weekdays et c.</short>
 Outputs all languages partially supported by roxen for writing
 numbers, weekdays et c (for example for the number and date tags).
 </p>
</desc>

 <ex><emit source='known-langs' sort='englishname'>
  4711 in &_.englishname;: <number lang='&_.id;' num='4711'/><br />
</emit></ex>",
			([
			  "&_.id;":#"<desc type='entity'>
 <p>Prints the three-character ISO 639-2 id of the language, for
 example \"eng\" for english and \"deu\" for german.</p>
</desc>",
			  "&_.name;":#"<desc type='entity'>
 <p>The name of the language in the language itself, for example
 \"franais\" for french.</p>
</desc>",
			  "&_.englishname;":#"<desc type='entity'>
 <p>The name of the language in English.</p>
</desc>",
			]) }),

#if 0
  // This applies to the old rxml 1.x parser. The doc for this tag is
  // here only for historical interest, to serve as a monument over
  // the quoting horrors that do_output_tag and its likes incurred.
  "recursive-output": #"\
<desc type='cont'>
  <p>This container provides a way to implement recursive output,
  which is mainly useful when you want to create arbitrarily nested
  trees from some external data, e.g. an SQL database. Put simply, the
  <tag>recurse</tag> tag is replaced by everything inside and
  including the <tag>recursive-output</tag> container. Although simple
  in theory, it tends to get a little bit messy in practice.</p>

  <p>To make it work you have to pay some attention to the parsing
  order of the involved tags. After the <tag>recursive-output</tag>
  container have replaced every <tag>recurse</tag> with itself, the
  whole thing is parsed again. Therefore, to make it terminate, you
  must always put the <tag>recurse</tag> inside a conditional
  container (typically an <tag>if</tag>) that does not preparse its
  contents.</p>

  <p>So far so good, but you'll almost always want to use some sort of
  output container, e.g. <tag>formoutput</tag> or
  <tag>sqloutput</tag>, together with this tag, which makes it
  slightly more complex due to the necessary treatment of the quote
  characters. Since the contents of <tag>recursive-output</tag> is
  expanded two levels at any time, each level needs its own set of
  quotes. To accomplish this, <tag>recursive-output</tag> can rotate
  two quote sets which are specified by the `inside' and `outside'
  arguments. Each time a <tag>recurse</tag> is replaced, every string
  in the `inside' set is replaced by the string in the corresponding
  position in the `outside' set, then the two sets trade places. Thus,
  you should put all quote characters you use inside
  <tag>recursive-output</tag> in the `inside' set and some other
  characters that doesn't clash with anything in the `outside' set.
  You might also have to quote the quote characters when writing these
  sets, which is done by doubling them.</p>
</desc>

<attr name='inside' value='string,...'><p>
  The list of quotes `inside' the container, to be replaced in with the
  `outside' set in every other round of recursion.</p>
</attr>

<attr name='outside' value='string,...'><p>
  The list of quotes `outside' the container, to be replaced in with
  the `inside' set in every other round of recursion.</p>
</attr>

<attr name='multisep' value='separator'><p>
  The given value is used as the separator between the strings in the
  two sets. It defaults to ','.</p>
</attr>

<attr name='limit' value='number'><p>
  Specifies the maximum nesting depth. As a safeguard it defaults to
  100.</p>
</attr>",
#endif

]);
#endif
